/* Modified by Ian Miller and Waseh Ahmad
 * File:   main.c
 * Author: Matthew Watkins & Syed Tahmid Mahbub
 * Main file for light controller module of the smart home lighting system
 * Created on October 19, 2014
 */


//IMPORTANT: Include config.h, then protothreads, then everything else
#include "config.h"
#include "pt_cornell_1_2.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plib.h"
#include "xc.h"
#include "adc_intf.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include "esp8266.h"
#include "TouchScreen.h"


#define SYS_FREQ 40000000
#define IP_GATEWAY "192.168.173.1"
#define NETMASK "255.255.255.0"
#define SERVER_IP "192.168.173.184"
#define CONTROLLER_IP "192.168.173.75"
#define CONTROLLER_ID "0"

#define WEB_SIGNAL "WEB_COMMAND"
#define MOTOR_SIGNAL "S"
#define WEB_INCR "INCR"
#define WEB_DECR "DECR"
#define INCREASE_COMMAND "I"
#define DECREASE_COMMAND "D"
#define STOP_COMMAND "S"
#define STOP_UNTARNISHED "+++"

//ambient light define
#define AMBIENT_LIGHT_WAIT_MSEC 200
#define AMBIENT_LIGHT_THRESH 5 //threshold on both sides of actual light to prevent noise from changing the signal
#define MAX_BRIGHTNESS 100


//buttons on touch screen
#define START_X_INCR 0
#define START_Y_INCR 0
#define BUT_WIDTH 150
#define BUT_HEIGHT 40
#define START_X_DECR START_X_INCR
#define START_Y_DECR 110

#define SCREEN_SCALE_X (0x666) //0.4 in fixed point (24.12)
#define SCREEN_SCALE_Y (0x4CC) //0.3 in fixed point (24.12)
#define INT_BIT 32
#define FRACT_BITS 8


static char lcd_buffer[20];

static struct pt pt_start_esp,pt_control_motor,pt_receive_commands,pt_ambient_light,pt_touch_interface;
static int act_ambient_light,set_brightness;

//booleans
static uint8_t stop_increasing,stop_decreasing,curr_increasing,curr_decreasing,manualControl,stopSent,prev_touched,release;



// <editor-fold defaultstate="collapsed" desc="Start ESP Thread which runs once">
//===============================================================================
//protothread to start the esp

char buffer[30];
static PT_THREAD (protothread_start_esp(struct pt *pt)){
    PT_BEGIN(pt);
    
    tft_setCursor(10, 10); 
    sprintf(buffer,"Restarting ESP");
    tft_fillRect(10, 10, 400, 100, ILI9341_BLACK);
    tft_writeString(buffer);
    
    Initialize_ESP8266();
    esp8266_restart();
    delay_ms(2000);
    do{
        tft_setCursor(10, 10);
        tft_setTextColor(ILI9341_WHITE);  tft_setTextSize(2);  
        sprintf(buffer,"Trying to start esp8266 \n module");
        tft_fillRect(10, 10, 400, 100, ILI9341_BLACK);
        tft_writeString(buffer);
        delay_ms(1500);    
    }while(!esp8266_isStarted());
    
    delay_ms(1500); 
    tft_setCursor(10, 10); 
    sprintf(buffer,"Connected to ESP");
    tft_fillRect(10, 10, 400, 100, ILI9341_BLACK);
    tft_writeString(buffer);
    
    //place esp8266 in station mode
    esp8266_mode(1);
    delay_ms(1500);
    //setup the ip address of the CIP
    esp8266_change_ip(SERVER_IP,IP_GATEWAY,NETMASK);
    //connect to network
    esp8266_connect("waseh","12345678");
    
    tft_setCursor(10, 10);
    sprintf(buffer,"Connected to Wifi");
    tft_fillRect(10, 10, 400, 100, ILI9341_BLACK);
    tft_writeString(buffer);
    
    delay_ms(1500);
    
    //set up in server mode
    esp8266_multiple_conn();
    esp8266_server_mode("80");
    
    PT_END(pt);
}
// </editor-fold>

//================================================================================
////protothread to receive commands from the web interface
static char receive_buffer[60];
static uint8_t state;
static uint16_t receive_size;
static PT_THREAD (protothread_receive_commands(struct pt *pt)){
    PT_BEGIN(pt);
    
    while(1){
        
        
        do{
            PT_YIELD_UNTIL(pt, UARTReceivedDataIsAvailable(UART2));//yield until data is available to be received
            //check whether +IPD is received
            switch(state){
                case 0:
                    if(UARTGetDataByte(UART2)=='+') state = 1;
                    break;
                case 1:
                    if(UARTGetDataByte(UART2)=='I') state = 2;
                    else state = 0;
                    break;
                case 2:
                    if(UARTGetDataByte(UART2)=='P') state = 3;
                    else state = 0;
                    break;
                case 3:
                    if(UARTGetDataByte(UART2)=='D') state = 4;
                    else state = 0;
                    break;
            }
        }while(state !=4);
        state = 0;
                
        //receive the remaining characters in the buffer
        esp8266_receive_all(receive_buffer,receive_size);
        //parse buffer to look for pin
        char *ret;
        ret = strstr(receive_buffer,WEB_SIGNAL);
        
        if(ret !=NULL){
            manualControl=1;
            //send appropriate motor command to the p
            if(strstr(receive_buffer,"INCR")){
                esp8266_send(INCREASE_COMMAND,CONTROLLER_ID);
            }else if(strstr(receive_buffer,"DECR")){
                esp8266_send(DECREASE_COMMAND,CONTROLLER_ID);
            }else{
                esp8266_send(STOP_COMMAND,CONTROLLER_ID);
            }
            //close the connection using AT+CIPCLOSe (for loop through all connections except 0)
            PT_YIELD_TIME_msec(1500);
            _esp8266_closeConnection("5");//close all connection. the other esp should reconnect automatically
        }else{
            ret = strstr(receive_buffer,MOTOR_SIGNAL);
            if(ret!=NULL){
                //stop sending signals
                if(curr_increasing){
                    //mPORTAToggleBits(BIT_2);//toggle other LED
                    if(act_ambient_light <(set_brightness -AMBIENT_LIGHT_THRESH)){
                        mPORTBSetBits(BIT_7);//turn on LED
                    }
                    stop_increasing=1;
                    stop_decreasing=0;
                }else if(curr_decreasing){
                    stop_decreasing=1;
                    stop_increasing=0;
                }else{//if it was not doing either and received the command, give precedence to the motor control
                    manualControl = 1;
                    stopSent = 1;
                    release = 1;
                    stop_increasing=0;
                    stop_decreasing=0;
                }
            }
            
        }
        //check for command from the MOtor controller
        
        PT_YIELD_TIME_msec(20);
    }
    
    PT_END(pt);
}


///================================================================================
//Thread to control the motor by telling it to increase brightness or decrease brightness
static PT_THREAD (protothread_control_motor(struct pt *pt)){
    PT_BEGIN(pt);
    
    while(1){
        //check for manual override first
        if(!manualControl){
            //read from the ambient light sensor and see if brightness level is within range (+-100 out of 1023)
            if(act_ambient_light <(set_brightness -AMBIENT_LIGHT_THRESH) && !curr_increasing && !stop_increasing){//if increase command hasnt been sent yet
                //command to motor controller to open shades i.e. decrease brightness
                if(esp8266_send(INCREASE_COMMAND,CONTROLLER_ID)){//if condition makes it so that if there is an error, it will keep sending the same command
                    curr_increasing = 1;
                    curr_decreasing=0;  
                    stop_decreasing = 0;//can reset this now as increase command has been sent
                }

            }else if(act_ambient_light >(set_brightness +AMBIENT_LIGHT_THRESH) && !curr_decreasing && !stop_decreasing){//if decrease command hasn't been set yet
                //command to close shades i.e. increase brightness
                //turn off LED first
                mPORTBClearBits(BIT_7);
                if(esp8266_send(DECREASE_COMMAND,CONTROLLER_ID)){
                    curr_decreasing = 1;
                    curr_increasing=0;
                    stop_increasing = 0;//can reset this now as decrease command has been sent
                }
            }else if(act_ambient_light >(set_brightness -AMBIENT_LIGHT_THRESH) && act_ambient_light <(set_brightness +AMBIENT_LIGHT_THRESH) && (curr_increasing||curr_decreasing)) {
                //stop moving i.e. send a stop command or stop sending a command
                if(esp8266_send(STOP_COMMAND,CONTROLLER_ID)){
                    PT_YIELD_TIME_msec(1500);
                    curr_increasing=0;
                    curr_decreasing=0;
                }
            }
            if(stop_increasing || stop_decreasing){
                curr_increasing=0;
                curr_decreasing=0;
            }
        }
        else{
            curr_increasing = 0;
            curr_decreasing = 0;
            stop_increasing = 0;
            stop_decreasing = 0;
        }
        
        
        
        // if current light is less than brightness leve
        
        PT_YIELD_TIME_msec(200);
    }
    
    PT_END(pt);
}

///================================================================================
//protothread to constantly read from ambient light sensor
static PT_THREAD (protothread_ambient_light(struct pt *pt)){
    PT_BEGIN(pt);
    
    while(1){        
        act_ambient_light = analogRead(5)/10;//read from analog pin 5 Pin#7 
        
        //print the ambient light percentage
        tft_fillRect(150, 10, 200, 200, ILI9341_BLACK);
        sprintf(lcd_buffer,"AMBIENT LIGHT");
        tft_setCursor(150, 10);
        tft_writeString(lcd_buffer);
        sprintf(lcd_buffer," %d %%",act_ambient_light);
        tft_setCursor(150, 30);
        tft_writeString(lcd_buffer);
        //print the set brightness percentage
        sprintf(lcd_buffer,"SET BRIGHTNESS");
        tft_setCursor(150, 50);
        tft_writeString(lcd_buffer);
        sprintf(lcd_buffer," %d %%",set_brightness);
        tft_setCursor(150, 70);
        tft_writeString(lcd_buffer);
        PT_YIELD_TIME_msec(AMBIENT_LIGHT_WAIT_MSEC);
        
    }
    
    PT_END(pt);
}



///================================================================================
//method to draw the buttons
static char label_buffer[30];
void drawButtons(){

        //Increase brightness;
        tft_fillRect(0, 0, BUT_WIDTH, BUT_HEIGHT, ILI9341_GREEN);
        //Decrease brightness
        tft_fillRect(0, BUT_HEIGHT, BUT_WIDTH, BUT_HEIGHT, ILI9341_RED);
        //manual increase button
        tft_fillRect(0, BUT_HEIGHT<<1, BUT_WIDTH, BUT_HEIGHT, ILI9341_BLUE);     
        //manual decrease button
        tft_fillRect(0, BUT_HEIGHT+(BUT_HEIGHT<<1), BUT_WIDTH, BUT_HEIGHT, ILI9341_PURPLE);

        
        //LED
        tft_fillRect(0, (BUT_HEIGHT<<2), BUT_WIDTH, BUT_HEIGHT, ILI9341_WHITE);
        
        //the +++ button
        tft_fillRect(0, BUT_HEIGHT+(BUT_HEIGHT<<2), BUT_WIDTH, BUT_HEIGHT, ILI9341_ORANGE);
        
        //close all connections button
        tft_fillRect(151, (BUT_HEIGHT+(BUT_HEIGHT<<2)), BUT_WIDTH, BUT_HEIGHT, ILI9341_PINK);
        
        //place labels for all buttons
        tft_setCursor(0, 0);
        tft_setTextSize(3);
        tft_setTextColor(ILI9341_BLACK);
        sprintf(label_buffer,"INCR BR.\n\nDECR BR.");//first two
        tft_writeString(label_buffer);
        
        tft_setTextColor(ILI9341_WHITE);
        sprintf(label_buffer,"\n\nINCR\nDECR\n\n");
        tft_writeString(label_buffer);
        
        tft_setTextColor(ILI9341_BLACK);
        sprintf(label_buffer,"LIGHT\n\nSTOP");
        tft_writeString(label_buffer);
        
        tft_setCursor(150,210);
        tft_setTextColor(ILI9341_WHITE);
        sprintf(label_buffer,"No Conn.");
        tft_writeString(label_buffer);
        
        //set for next drawings
        tft_setTextSize(2);
        tft_setTextColor(ILI9341_WHITE);

}


///================================================================================
//protothread to interact with the touch screen interface to decrease or increase brightness


static uint32_t pixel_x,pixel_y;
static PT_THREAD (protothread_touch_interface(struct pt *pt)){
    PT_BEGIN(pt);
    
    while(1){
               
        
        struct TSPoint p;
        p.x = 0;
        p.y = 0;
        p.z = 0;
        getPoint(&p);
       
        //The touch coordinates are flipped.  Go figure
        pixel_x = 320-(((((uint32_t)(p.y-120))<<FRACT_BITS)*SCREEN_SCALE_X)>>(FRACT_BITS+12));
        pixel_y = 240-(((((uint32_t)(p.x-100))<<FRACT_BITS)*SCREEN_SCALE_Y)>>(FRACT_BITS+12));
        
        if(p.z > 50 && p.z < 800){
            
            //which area does press occur
            if(pixel_x<BUT_WIDTH){
                if(pixel_y<(BUT_HEIGHT) && set_brightness<=MAX_BRIGHTNESS){
                    //increase button pressed
                    set_brightness=set_brightness+1;//might need to cap this
                    //turn off manual control first
                    manualControl = 0;
                    
                }else if(pixel_y<(BUT_HEIGHT<<1) && set_brightness>0){
                    set_brightness=set_brightness-1; 
                    //turn off manual control first
                    manualControl = 0;
                    
                }else if(pixel_y<(BUT_HEIGHT+(BUT_HEIGHT<<1)) && stopSent && !release && !prev_touched){//release makes it so that when the motor reaches end while held down, a stop wotn be sent
                    manualControl=1;
                    esp8266_send(INCREASE_COMMAND,CONTROLLER_ID);
                    stopSent=0;
                }else if(pixel_y<(BUT_HEIGHT<<2) && stopSent && !release && !prev_touched){
                    manualControl=1;
                    esp8266_send(DECREASE_COMMAND,CONTROLLER_ID);
                    stopSent=0;
                }else if(pixel_y<(BUT_HEIGHT+(BUT_HEIGHT<<2)) && !prev_touched){
                    manualControl=1;
                    mPORTBToggleBits(BIT_7);
                    
                }else if(pixel_y<((BUT_HEIGHT<<1)+(BUT_HEIGHT<<2)) && !prev_touched){//send a stop 
                    stopSent=1;
                    esp8266_send(STOP_COMMAND,CONTROLLER_ID);
                }
                
            }
            else{
                if(pixel_y>(BUT_HEIGHT+ (BUT_HEIGHT<<2)) && !prev_touched){
                    _esp8266_closeConnection("5");//close all connections
                }   
            }
            
            prev_touched = 1;
        }else{
            release = 0;
            prev_touched = 0;
            
            if(manualControl && !stopSent){//after the finger is lifted  
                stopSent=1;
            }
        }
    
        
        
        PT_YIELD_TIME_msec(200);
    }
    
    PT_END(pt);
}
///================================================================================


int main(int argc, char** argv) {
    
    SYSTEMConfigPerformance(SYS_FREQ);

    CM1CON = 0; CM2CON = 0; ANSELA = 0; ANSELB = 0;
    
    configureADC();
    
    //initialize screen
    tft_init_hw();
    tft_begin();
    tft_setRotation(1); 
    tft_fillScreen(ILI9341_BLACK);  
    
    PT_setup();
    
    //setup threads
    PT_INIT(&pt_start_esp);
    PT_INIT(&pt_control_motor);
    PT_INIT(&pt_receive_commands);
    PT_INIT(&pt_ambient_light);
    PT_INIT(&pt_touch_interface);
    
    
    //pin out
    PORTSetPinsAnalogIn(IOPORT_B,BIT_3);
    
    mPORTBSetPinsDigitalOut(BIT_7);
    mPORTASetPinsDigitalOut(BIT_2);
    
    
    //initialize variables
    set_brightness = 50;
    receive_size = 40;
    curr_increasing=0;
    curr_decreasing=0;
    stop_increasing=0;
    stop_decreasing=0;
    state = 0;
    manualControl = 1;
    stopSent = 1;
    release = 0;
    
    
    //Global Interrupt Enable
    INTEnableSystemMultiVectoredInt();  
    
    //let this thread run once to start the esp
    PT_SCHEDULE(protothread_start_esp(&pt_start_esp));
    
    tft_fillScreen(ILI9341_BLACK); 
    //draw the buttons on the screen
    drawButtons();
    
    //Let's thread!
    while (1){
        PT_SCHEDULE(protothread_control_motor(&pt_control_motor));
        PT_SCHEDULE(protothread_receive_commands(&pt_receive_commands));
        PT_SCHEDULE(protothread_ambient_light(&pt_ambient_light));
        PT_SCHEDULE(protothread_touch_interface(&pt_touch_interface));
        
    }
}
